"""Object node item."""

from gaphor import UML
from gaphor.core.modeling.properties import attribute
from gaphor.diagram.presentation import ElementPresentation, Named
from gaphor.diagram.shapes import Box, IconBox, Text, draw_border
from gaphor.diagram.support import represents
from gaphor.i18n import i18nize
from gaphor.UML.recipes import stereotypes_str

DEFAULT_UPPER_BOUND = "*"

ORDERING_TEXT = {
    "unordered": i18nize("unordered"),
    "ordered": i18nize("ordered"),
    "LIFO": i18nize("LIFO"),
    "FIFO": i18nize("FIFO"),
}


@represents(UML.ObjectNode)
class ObjectNodeItem(Named, ElementPresentation):
    """Representation of object node. Object node is ordered and has upper
    bound specification.

    Ordering information can be hidden by user.
    """

    def __init__(self, diagram, id=None):
        super().__init__(
            diagram,
            id,
            shape=IconBox(
                Box(
                    Text(
                        text=lambda: stereotypes_str(self.subject),
                    ),
                    Text(text=lambda: self.subject.name or ""),
                    style={
                        "padding": (4, 12, 4, 12),
                    },
                    draw=draw_border,
                ),
                Text(
                    text=lambda: self.subject.upperBound
                    not in (None, "", DEFAULT_UPPER_BOUND)
                    and f'{{ {diagram.gettext("upperBound")} = {self.subject.upperBound} }}'
                    or ""
                ),
                Text(
                    text=lambda: self.show_ordering
                    and self.subject.ordering
                    and f'{{ {diagram.gettext("ordering")} = {diagram.gettext(ORDERING_TEXT.get(self.subject.ordering))} }}'
                    or ""
                ),
            ),
            width=50,
            height=30,
        )

        self.watch("subject[NamedElement].name")
        self.watch("subject.appliedStereotype.classifier.name")
        self.watch("subject[ObjectNode].upperBound")
        self.watch("subject[ObjectNode].ordering")
        self.watch("show_ordering")

    show_ordering: attribute[int] = attribute("show_ordering", int, default=False)

    def load(self, name, value):
        if name == "show-ordering":
            name = "show_ordering"
        super().load(name, value)



"""Interface item related connections.

The connectors implemented in this module check if connection is
possible to folded interface, see `gaphor.diagram.classes.interface`
documentation for details.
"""


from gaphor.diagram.connectors import Connector
from gaphor.UML.classes.classconnect import DependencyConnect
from gaphor.UML.classes.dependency import DependencyItem
from gaphor.UML.classes.interface import InterfaceItem


@Connector.register(InterfaceItem, DependencyItem)
class DependencyInterfaceConnect(DependencyConnect):
    """Connect interface item with dependency item."""

    element: InterfaceItem

    def connect(self, handle, port):
        """Dependency item is changed to draw in solid mode, when connected to
        folded interface."""
        super().connect(handle, port)
        line = self.line
        # connecting to the interface, which is supplier - assuming usage
        # dependency
        if handle is line.head:
            # change interface side even when it is unfolded, this way
            # required interface will be rotated properly when folded by
            # user
            self.element.side = port.side
            self.element.request_update()
            self.line.request_update()
            self.element.update_shapes()

    def disconnect(self, handle):
        """If dependency item is no longer connected to an interface, then draw
        it in non-solid mode.

        Interface's folded mode changes to provided (ball) notation.
        """
        super().disconnect(handle)
        if handle is self.line.head:
            self.element.request_update()
            self.line.request_update()



from gaphor.diagram.support import represents
from gaphor.SysML import sysml
from gaphor.SysML.relationships import DirectedRelationshipPropertyPathItem


@represents(
    sysml.Satisfy, head=sysml.Satisfy.sourceContext, tail=sysml.Satisfy.targetContext
)
class SatisfyItem(DirectedRelationshipPropertyPathItem):
    @property
    def relation_type(self):
        return self.diagram.gettext("satisfy")


@represents(
    sysml.DeriveReqt,
    head=sysml.DeriveReqt.sourceContext,
    tail=sysml.DeriveReqt.targetContext,
)
class DeriveReqtItem(DirectedRelationshipPropertyPathItem):
    @property
    def relation_type(self):
        return self.diagram.gettext("deriveReqt")


@represents(sysml.Trace, head=sysml.Trace.sourceContext, tail=sysml.Trace.targetContext)
class TraceItem(DirectedRelationshipPropertyPathItem):
    @property
    def relation_type(self):
        return self.diagram.gettext("trace")


@represents(
    sysml.Verify, head=sysml.Verify.sourceContext, tail=sysml.Verify.targetContext
)
class VerifyItem(DirectedRelationshipPropertyPathItem):
    @property
    def relation_type(self):
        return self.diagram.gettext("verify")


@represents(
    sysml.Refine, head=sysml.Refine.sourceContext, tail=sysml.Refine.targetContext
)
class RefineItem(DirectedRelationshipPropertyPathItem):
    @property
    def relation_type(self):
        return self.diagram.gettext("refine")



from gi.repository import Gdk, Gtk

from gaphor.core.modeling import Diagram
from gaphor.diagram.event import DiagramOpened
from gaphor.diagram.instanteditors import instant_editor


def text_edit_tools(event_manager):
    key_tool = Gtk.EventControllerKey.new()
    click_tool = Gtk.GestureClick.new()

    key_tool.connect("key-pressed", on_key_pressed, event_manager)
    click_tool.connect("released", on_double_click, event_manager)
    return key_tool, click_tool


def on_key_pressed(controller, keyval, keycode, state, event_manager):
    view = controller.get_widget()
    item = view.selection.hovered_item
    if item and keyval == Gdk.KEY_F2:
        return instant_editor(item, view, event_manager)


def on_double_click(gesture, n_press, x, y, event_manager):
    view = gesture.get_widget()
    item = view.selection.hovered_item
    if not item or n_press != 2:
        return

    if isinstance(item.subject, Diagram):
        event_manager.handle(DiagramOpened(item.subject))
    else:
        ix, iy = view.get_matrix_v2i(item).transform_point(x, y)
        return instant_editor(item, view, event_manager, (ix, iy))



from typing import Union

from gaphor.C4Model import c4model
from gaphor.core import transactional
from gaphor.diagram.propertypages import (
    PropertyPageBase,
    PropertyPages,
    handler_blocking,
    new_resource_builder,
    unsubscribe_all_on_destroy,
)

new_builder = new_resource_builder("gaphor.C4Model")


@PropertyPages.register(c4model.C4Container)
@PropertyPages.register(c4model.C4Person)
class DescriptionPropertyPage(PropertyPageBase):
    order = 14

    def __init__(self, subject: Union[c4model.C4Container, c4model.C4Person]):
        super().__init__()
        assert subject
        self.subject = subject
        self.watcher = subject.watcher()

    def construct(self):
        builder = new_builder(
            "description-editor",
            "description-text-buffer",
        )
        subject = self.subject

        description = builder.get_object("description")

        buffer = builder.get_object("description-text-buffer")
        if subject.description:
            buffer.set_text(subject.description)

        @handler_blocking(buffer, "changed", self._on_description_changed)
        def text_handler(event):
            if not description.props.has_focus:
                buffer.set_text(event.new_value)

        self.watcher.watch("description", text_handler)

        return unsubscribe_all_on_destroy(
            builder.get_object("description-editor"), self.watcher
        )

    @transactional
    def _on_description_changed(self, buffer):
        self.subject.description = buffer.get_text(
            buffer.get_start_iter(), buffer.get_end_iter(), False
        )


@PropertyPages.register(c4model.C4Container)
class TechnologyPropertyPage(PropertyPageBase):
    order = 15

    def __init__(self, subject: c4model.C4Container):
        super().__init__()
        assert subject
        self.subject = subject

    def construct(self):
        builder = new_builder(
            "technology-editor",
            signals={
                "technology-changed": (self._on_technology_changed,),
            },
        )
        subject = self.subject

        technology = builder.get_object("technology")
        technology.set_text(subject.technology or "")

        return builder.get_object("technology-editor")

    @transactional
    def _on_technology_changed(self, entry):
        self.subject.technology = entry.get_text()



import functools
import importlib.metadata
import inspect
import logging
from typing import Dict, TypeVar

T = TypeVar("T")

logger = logging.getLogger(__name__)


def initialize(scope, services=None, **known_services: T) -> Dict[str, T]:
    return init_entry_points(load_entry_points(scope, services), **known_services)


@functools.lru_cache(maxsize=4)
def list_entry_points(group):
    return importlib.metadata.entry_points(group=group)


def load_entry_points(scope, services=None) -> Dict[str, type]:
    """Load services from resources."""
    uninitialized_services = {}
    for ep in list_entry_points(scope):
        cls = ep.load()
        if not services or ep.name in services:
            logger.debug(f'found entry point "{scope}.{ep.name}"')
            uninitialized_services[ep.name] = cls
    return uninitialized_services


def init_entry_points(
    uninitialized_services: Dict[str, type[T]], **known_services: T
) -> Dict[str, T]:
    """Instantiate service definitions, taking into account dependencies
    defined in the constructor.

    Given a dictionary `{name: service-class}`, return a map `{name:
    service-instance}`.
    """
    ready: Dict[str, T] = known_services.copy()

    def pop(name):
        try:
            return uninitialized_services.pop(name)
        except KeyError:
            return None

    def init(name, cls):
        kwargs = {}
        for param_name, param in inspect.signature(cls).parameters.items():
            if param_name not in ready:
                depcls = pop(param_name)
                if depcls:
                    kwargs[param_name] = init(param_name, depcls)
                elif param.default is inspect.Parameter.empty:
                    logger.warn(
                        "Entrypoint %s parameter %s does not reference a resolved dependency",
                        name,
                        param_name,
                    )
            else:
                kwargs[param_name] = ready[param_name]
        srv = cls(**kwargs)
        ready[name] = srv
        return srv

    while uninitialized_services:
        name = next(iter(uninitialized_services.keys()))
        cls = pop(name)
        init(name, cls)

    return ready



"""The action definition for the RAAML toolbox."""
from gaphor.diagram.diagramtoolbox import (
    DiagramType,
    DiagramTypes,
    ElementCreateInfo,
    ToolboxDefinition,
    general_tools,
)
from gaphor.i18n import i18nize
from gaphor.RAAML.fta.ftatoolbox import fta
from gaphor.RAAML.raaml import Hazard, Loss, Situation, TopEvent
from gaphor.RAAML.stpa.stpatoolbox import stpa
from gaphor.UML.uml import Package

raaml_toolbox_actions: ToolboxDefinition = (
    general_tools,
    fta,
    stpa,
)

raaml_diagram_types: DiagramTypes = (
    DiagramType("fta", i18nize("New FTA Diagram"), (fta,)),
    DiagramType("stpa", i18nize("New STPA Diagram"), (stpa,)),
)

raaml_element_types = (
    ElementCreateInfo("topevent", i18nize("New Top Event"), TopEvent, (Package,)),
    ElementCreateInfo("loss", i18nize("New Loss"), Loss, (Package,)),
    ElementCreateInfo("hazard", i18nize("New Hazard"), Hazard, (Package,)),
    ElementCreateInfo("situation", i18nize("New Situation"), Situation, (Package,)),
)



from gi.repository import Gtk

from gaphor import UML
from gaphor.core import transactional
from gaphor.diagram.propertypages import (
    PropertyPageBase,
    PropertyPages,
    handler_blocking,
    new_resource_builder,
    unsubscribe_all_on_destroy,
)

new_builder = new_resource_builder("gaphor.UML.profiles")


def _issubclass(c, b):
    try:
        return issubclass(c, b)
    except TypeError:
        return False


@PropertyPages.register(UML.Class)
class MetaclassPropertyPage(PropertyPageBase):
    """Adapter which shows a property page for a class view.

    Also handles metaclasses.
    """

    order = 10

    subject: UML.Class

    CLASSES = sorted(
        c
        for c in dir(UML)
        if _issubclass(getattr(UML, c), UML.Element) and c != "Stereotype"
    )

    def __init__(self, subject: UML.Class):
        self.subject = subject
        self.watcher = subject.watcher()

    def construct(self):
        if not UML.recipes.is_metaclass(self.subject):
            return

        builder = new_builder(
            "metaclass-editor",
        )

        dropdown = builder.get_object("metaclass-dropdown")
        dropdown.set_model(Gtk.StringList.new(self.CLASSES))
        if self.subject and self.subject.name in self.CLASSES:
            dropdown.set_selected(self.CLASSES.index(self.subject.name))

        @handler_blocking(dropdown, "notify::selected-item", self._on_name_changed)
        def handler(event):
            if event.element is self.subject and event.new_value in self.CLASSES:
                dropdown.set_selected(self.CLASSES.index(self.subject.name))

        self.watcher.watch("name", handler)

        return unsubscribe_all_on_destroy(
            builder.get_object("metaclass-editor"), self.watcher
        )

    @transactional
    def _on_name_changed(self, dropdown, _pspec):
        self.subject.name = dropdown.get_selected_item().get_string()



"""Base code for presentation elements."""

from __future__ import annotations

import ast
import logging
import re
from typing import TYPE_CHECKING, Generic, TypeVar

from gaphas.item import Matrices

from gaphor.core.modeling.element import Element, Id, UnlinkEvent
from gaphor.core.modeling.event import RevertibleEvent
from gaphor.core.modeling.properties import relation_many, relation_one

if TYPE_CHECKING:
    from gaphor.core.modeling.diagram import Diagram

S = TypeVar("S", bound=Element)


log = logging.getLogger(__name__)


def literal_eval(value: str):
    return ast.literal_eval(re.sub("\r|\n", "", value))


class Presentation(Matrices, Element, Generic[S]):
    """This presentation is used to link the behaviors of
    `gaphor.core.modeling` and `gaphas.Item`."""

    def __init__(self, diagram: Diagram, id: Id | None = None) -> None:
        super().__init__(id=id, model=diagram.model)
        self.diagram = diagram
        self._original_diagram: Diagram | None = diagram

        def update(event):
            if self.diagram:
                self.diagram.request_update(self)

        self._watcher = self.watcher(default_handler=update)
        self.watch("subject")
        self.watch("children")
        self.watch("diagram", self._on_diagram_changed)
        self.watch("parent", self._on_parent_changed)
        self.matrix.add_handler(self._on_matrix_changed)

    subject: relation_one[S]
    diagram: relation_one[Diagram]
    parent: relation_one[Presentation]
    children: relation_many[Presentation]

    def request_update(self):
        if self.diagram:
            self.diagram.request_update(self)

    def watch(self, path, handler=None):
        """Watch a certain path of elements starting with the DiagramItem. The
        handler is optional and will default to a simple self.request_update().

        Watches should be set in the constructor, so they can be registered
        and unregistered in one shot.

        This interface is fluent(returns self).
        """
        self._watcher.watch(path, handler)
        return self

    def change_parent(self, new_parent):
        """Change the parent and update the item's matrix so the item visually
        remains in the same place."""
        old_parent = self.parent
        if new_parent is old_parent:
            return

        self.parent = new_parent
        m = self.matrix
        if old_parent:
            m = m * old_parent.matrix_i2c
        if new_parent:
            m = m * new_parent.matrix_i2c.inverse()
        self.matrix.set(*m)

    def css_nodes(self):
        return ()

    def load(self, name, value):
        if name == "matrix":
            self.matrix.set(*literal_eval(value))
        elif name == "parent":
            if self.parent and self.parent is not value:
                raise ValueError(f"Parent can not be set twice on {self}")
            super().load(name, value)
            self.parent.matrix_i2c.add_handler(self._on_matrix_changed)
            self._on_matrix_changed(None, ())
        else:
            super().load(name, value)

    def inner_unlink(self, _unlink_event: UnlinkEvent) -> None:
        self._watcher.unsubscribe_all()
        self.matrix.remove_handler(self._on_matrix_changed)

        if self.parent:
            self.parent.matrix_i2c.remove_handler(self._on_matrix_changed)

        if diagram := self._original_diagram:
            connecting_items = [
                cinfo.item
                for cinfo in diagram.connections.get_connections(connected=self)
            ]

            diagram.connections.remove_connections_to_item(self)
            for item in connecting_items:
                item.unlink()
            self._original_diagram = None
            super().inner_unlink(UnlinkEvent(self, diagram=diagram))

    def _on_diagram_changed(self, event):
        new_value = event.new_value
        if new_value and new_value is not self._original_diagram:
            self.diagram = self._original_diagram
            raise ValueError("Can't change diagram of a presentation")

    def _on_parent_changed(self, event):
        if old_parent := event.old_value:
            old_parent.matrix_i2c.remove_handler(self._on_matrix_changed)

        if new_parent := event.new_value:
            new_parent.matrix_i2c.add_handler(self._on_matrix_changed)

        self._on_matrix_changed(None, ())

    def _on_matrix_changed(self, matrix, old_value):
        if self.parent:
            self.matrix_i2c.set(*(self.matrix * self.parent.matrix_i2c))
        else:
            self.matrix_i2c.set(*self.matrix)
        self.request_update()
        if matrix is self.matrix:
            self.handle(MatrixUpdated(self, old_value))


class MatrixUpdated(RevertibleEvent):
    def __init__(self, element, old_value):
        super().__init__(element)
        self.old_value = old_value

    def revert(self, target):
        target.matrix.set(*self.old_value)



"""
CommentLine -- A line that connects a comment to another model element.

"""

from gaphor.diagram.connectors import Connector
from gaphor.diagram.presentation import LinePresentation


class CommentLineItem(LinePresentation):
    def __init__(self, diagram, id=None):
        super().__init__(diagram, id, style={"dash-style": (7.0, 5.0)})

    def unlink(self):
        c1 = self._connections.get_connection(self.head)
        c2 = self._connections.get_connection(self.tail)
        if c1 and c2:
            adapter = Connector(c1.connected, self)
            adapter.disconnect(self.head)
        super().unlink()



from gaphor.UML.deployments import (  # noqa: F401
    connectorconnect,
    copypaste,
    deploymentsgroup,
)
from gaphor.UML.deployments.artifact import ArtifactItem
from gaphor.UML.deployments.connector import ConnectorItem
from gaphor.UML.deployments.node import NodeItem

__all__ = ["ArtifactItem", "ConnectorItem", "NodeItem"]



from gaphor.diagram.support import represents
from gaphor.SysML.relationships import DirectedRelationshipPropertyPathItem
from gaphor.SysML.sysml import Allocate


@represents(Allocate, head=Allocate.sourceContext, tail=Allocate.targetContext)
class AllocateItem(DirectedRelationshipPropertyPathItem):
    @property
    def relation_type(self):
        return self.diagram.gettext("allocate")



# ruff: noqa: F401

import gaphor.SysML.blocks.connectors
import gaphor.SysML.blocks.datatype
import gaphor.SysML.blocks.group
from gaphor.SysML.blocks.block import BlockItem
from gaphor.SysML.blocks.interfaceblock import InterfaceBlockItem
from gaphor.SysML.blocks.property import PropertyItem
from gaphor.SysML.blocks.proxyport import ProxyPortItem



import itertools

from gaphor.diagram.copypaste import copy
from gaphor.UML import Association, Class, DataType, Enumeration, Interface, Operation
from gaphor.UML.copypaste import copy_named_element


@copy.register(Class)
@copy.register(DataType)
@copy.register(Interface)
def copy_class(element):
    yield element.id, copy_named_element(element)
    for feature in itertools.chain(
        element.ownedAttribute,
        element.ownedOperation,
    ):
        yield from copy(feature)


@copy.register
def copy_enumeration(element: Enumeration):
    yield element.id, copy_named_element(element)
    for feature in itertools.chain(
        element.ownedAttribute,
        element.ownedOperation,
        element.ownedLiteral,
    ):
        yield from copy(feature)


@copy.register
def copy_operation(element: Operation):
    yield element.id, copy_named_element(element)
    for feature in element.ownedParameter:
        yield from copy(feature)


@copy.register
def copy_association(element: Association):
    yield element.id, copy_named_element(element)
    for end in element.memberEnd:
        yield from copy(end)



"""The definition for the classes section of the toolbox."""

from gaphas.item import SE

from gaphor import UML
from gaphor.core import gettext
from gaphor.diagram.diagramtoolbox import ToolDef, ToolSection, new_item_factory
from gaphor.UML import diagramitems
from gaphor.UML.toolboxconfig import namespace_config


def composite_association_config(assoc_item: diagramitems.AssociationItem) -> None:
    assoc_item.preferred_aggregation = "composite"


def shared_association_config(assoc_item: diagramitems.AssociationItem) -> None:
    assoc_item.preferred_aggregation = "shared"


def direct_association_config(assoc_item: diagramitems.AssociationItem) -> None:
    assoc_item.preferred_tail_navigability = "navigable"


classes = ToolSection(
    gettext("Classes"),
    (
        ToolDef(
            "toolbox-class",
            gettext("Class"),
            "gaphor-class-symbolic",
            "c",
            new_item_factory(
                diagramitems.ClassItem, UML.Class, config_func=namespace_config
            ),
            handle_index=SE,
        ),
        ToolDef(
            "toolbox-interface",
            gettext("Interface"),
            "gaphor-interface-symbolic",
            "i",
            new_item_factory(
                diagramitems.InterfaceItem,
                UML.Interface,
                config_func=namespace_config,
            ),
            handle_index=SE,
        ),
        ToolDef(
            "toolbox-package",
            gettext("Package"),
            "gaphor-package-symbolic",
            "p",
            new_item_factory(
                diagramitems.PackageItem,
                UML.Package,
                config_func=namespace_config,
            ),
            handle_index=SE,
        ),
        ToolDef(
            "toolbox-component",
            gettext("Component"),
            "gaphor-component-symbolic",
            "o",
            new_item_factory(
                diagramitems.ComponentItem,
                UML.Component,
                config_func=namespace_config,
            ),
            handle_index=SE,
        ),
        ToolDef(
            "toolbox-containment",
            gettext("Containment"),
            "gaphor-containment-symbolic",
            "<Shift>C",
            new_item_factory(diagramitems.ContainmentItem),
        ),
        ToolDef(
            "toolbox-composite-association",
            gettext("Composite Association"),
            "gaphor-composite-association-symbolic",
            "<Shift>Z",
            new_item_factory(
                diagramitems.AssociationItem,
                config_func=composite_association_config,
            ),
        ),
        ToolDef(
            "toolbox-shared-association",
            gettext("Shared Association"),
            "gaphor-shared-association-symbolic",
            "<Shift>Q",
            new_item_factory(
                diagramitems.AssociationItem,
                config_func=shared_association_config,
            ),
        ),
        ToolDef(
            "toolbox-association",
            gettext("Association"),
            "gaphor-association-symbolic",
            "<Shift>A",
            new_item_factory(diagramitems.AssociationItem),
        ),
        ToolDef(
            "toolbox-direct-association",
            gettext("Direct Association"),
            "gaphor-direct-association-symbolic",
            None,
            new_item_factory(
                diagramitems.AssociationItem,
                config_func=direct_association_config,
            ),
        ),
        ToolDef(
            "toolbox-dependency",
            gettext("Dependency"),
            "gaphor-dependency-symbolic",
            "<Shift>D",
            new_item_factory(diagramitems.DependencyItem),
            handle_index=0,
        ),
        ToolDef(
            "toolbox-generalization",
            gettext("Generalization"),
            "gaphor-generalization-symbolic",
            "<Shift>G",
            new_item_factory(diagramitems.GeneralizationItem),
            handle_index=0,
        ),
        ToolDef(
            "toolbox-interface-realization",
            gettext("Interface Realization"),
            "gaphor-interface-realization-symbolic",
            "<Shift>I",
            new_item_factory(diagramitems.InterfaceRealizationItem),
        ),
        ToolDef(
            "toolbox-data-type",
            gettext("DataType"),
            "gaphor-data-type-symbolic",
            "<Shift>L",
            new_item_factory(
                diagramitems.DataTypeItem,
                UML.DataType,
                config_func=namespace_config,
            ),
            handle_index=SE,
        ),
        ToolDef(
            "toolbox-enumeration",
            gettext("Enumeration"),
            "gaphor-enumeration-symbolic",
            "<Shift>W",
            new_item_factory(
                diagramitems.EnumerationItem,
                UML.Enumeration,
                config_func=namespace_config,
            ),
            handle_index=SE,
        ),
        ToolDef(
            "toolbox-primitive",
            gettext("Primitive"),
            "gaphor-primitive-type-symbolic",
            "<Shift>H",
            new_item_factory(
                diagramitems.DataTypeItem,
                UML.PrimitiveType,
                config_func=namespace_config,
            ),
            handle_index=SE,
        ),
    ),
)



import logging

from gaphor import UML
from gaphor.core.modeling.properties import attribute
from gaphor.core.styling import JustifyContent
from gaphor.diagram.presentation import (
    Classified,
    ElementPresentation,
    from_package_str,
    text_name,
)
from gaphor.diagram.shapes import Box, Text, draw_border
from gaphor.diagram.support import represents
from gaphor.UML.classes.klass import (
    attribute_watches,
    attributes_compartment,
    operation_watches,
    operations_compartment,
)
from gaphor.UML.classes.stereotype import stereotype_compartments, stereotype_watches
from gaphor.UML.recipes import stereotypes_str

log = logging.getLogger(__name__)


@represents(UML.DataType)
@represents(UML.PrimitiveType)
class DataTypeItem(Classified, ElementPresentation[UML.DataType]):
    """This item visualizes a Data Type instance.

    A DataTypeItem contains two compartments:
    1. Attributes
    2. Operations
    """

    def __init__(self, diagram, id=None):
        super().__init__(diagram, id=id)

        self.watch("show_attributes", self.update_shapes).watch(
            "show_operations", self.update_shapes
        ).watch("subject[NamedElement].name").watch(
            "subject[NamedElement].namespace.name"
        )
        attribute_watches(self, "DataType")
        operation_watches(self, "DataType")
        stereotype_watches(self)

    show_stereotypes: attribute[int] = attribute("show_stereotypes", int)

    show_attributes: attribute[int] = attribute("show_attributes", int, default=True)

    show_operations: attribute[int] = attribute("show_operations", int, default=True)

    def additional_stereotypes(self):
        from gaphor.SysML import sysml

        if isinstance(self.subject, UML.PrimitiveType):
            return [self.diagram.gettext("primitive")]
        elif isinstance(self.subject, sysml.ValueType):
            return [self.diagram.gettext("valueType")]
        elif isinstance(self.subject, UML.DataType):
            return [self.diagram.gettext("dataType")]
        return ()

    def update_shapes(self, event=None):
        self.shape = Box(
            Box(
                Text(
                    text=lambda: stereotypes_str(
                        self.subject, self.additional_stereotypes()
                    ),
                ),
                text_name(self),
                Text(
                    text=lambda: from_package_str(self),
                    style={"font-size": "x-small"},
                ),
                style={"padding": (12, 4, 12, 4)},
            ),
            *(
                self.show_attributes
                and self.subject
                and [attributes_compartment(self.subject)]
                or []
            ),
            *(
                self.show_operations
                and self.subject
                and [operations_compartment(self.subject)]
                or []
            ),
            *(self.show_stereotypes and stereotype_compartments(self.subject) or []),
            style={
                "justify-content": JustifyContent.START,
            },
            draw=draw_border,
        )



import logging

from gaphor.abc import Service
from gaphor.entrypoint import list_entry_points

log = logging.getLogger(__name__)


class ModuleLoader(Service):
    def __init__(self):
        """Load modules defined in entrypoints `[gaphor.modules]`."""

        for ep in list_entry_points("gaphor.modules"):
            log.debug(f'found module entry point "{ep.name}"')
            ep.load()

    def shutdown(self):
        pass



from __future__ import annotations

import datetime
import logging
import sys
import textwrap
import time
from types import TracebackType

import better_exceptions
from exceptiongroup import BaseExceptionGroup
from gi.repository import Gtk

from gaphor.abc import ActionProvider
from gaphor.action import action
from gaphor.core.eventmanager import event_handler
from gaphor.event import Notification, SessionCreated
from gaphor.i18n import gettext, translated_ui_string
from gaphor.ui.abc import UIComponent
from gaphor.ui.selftest import system_information

better_exceptions.SUPPORTS_COLOR = False

log = logging.getLogger(__name__)

START_TIME = time.time()


def new_builder():
    builder = Gtk.Builder()
    builder.add_from_string(
        translated_ui_string("gaphor.plugins.errorreports", "errorreports.ui")
    )
    return builder


class ErrorReports(UIComponent, ActionProvider):
    def __init__(self, application, event_manager):
        self.application = application
        self.event_manager = event_manager
        self.window = None
        self.buffer = None
        self.exceptions: list[
            tuple[float, type[BaseException], BaseException, TracebackType | None]
        ] = []
        self._orig_excepthook = sys.excepthook
        sys.excepthook = self.excepthook
        event_manager.subscribe(self.on_session_created)

    def shutdown(self):
        super().shutdown()
        sys.excepthook = self._orig_excepthook
        self.event_manager.unsubscribe(self.on_session_created)

    @event_handler(SessionCreated)
    def on_session_created(self, event):
        session = event.session
        tools_menu = session.get_service("tools_menu")
        tools_menu.add_actions(self, scope="app")

    @action(name="app.error-reports-open", label=gettext("Error Reports"))
    def open(self):
        if self.window:
            self.window.present()
            return

        builder = new_builder()

        self.window = builder.get_object("error-reports")
        self.buffer = builder.get_object("buffer")

        self.window.connect("destroy", self.close)
        self.window.set_visible(True)

        self.update_text()

    def close(self, widget=None):
        if self.window:
            self.window.destroy()
            self.window = None
            self.buffer = None

    def update_text(self):
        buffer = self.buffer
        if not buffer:
            return

        buffer.delete(buffer.get_start_iter(), buffer.get_end_iter())

        buffer.insert(buffer.get_end_iter(), system_information())
        buffer.insert(buffer.get_end_iter(), "\n\n")

        if not self.exceptions:
            buffer.insert(
                buffer.get_end_iter(),
                gettext("No errors have been intercepted. We’re good."),
            )
            return

        buffer.insert(buffer.get_end_iter(), gettext("Errors:"))
        buffer.insert(buffer.get_end_iter(), "\n\n")

        def print_exception(tp, v, tb, depth=0):
            assert buffer
            for line in better_exceptions.format_exception(tp, v, tb):
                buffer.insert(
                    buffer.get_end_iter(), textwrap.indent(line, "  " * depth + "|")
                )
            if issubclass(tp, BaseExceptionGroup):
                for i, sub_exc in enumerate(v.exceptions):
                    buffer.insert(
                        buffer.get_end_iter(),
                        f"{'  ' * depth}{'└─┬──' if i == 0 else '  ├──'}────────────────────────────╌┄┈\n",
                    )
                    print_exception(
                        type(sub_exc), sub_exc, sub_exc.__traceback__, depth=depth + 1
                    )
                buffer.insert(
                    buffer.get_end_iter(),
                    f"{'  ' * (depth + 1)}└─────────────────────────────╌┄┈\n",
                )
            elif depth == 0:
                buffer.insert(
                    buffer.get_end_iter(),
                    "└─────────────────────────────╌┄┈\n",
                )

        for ts, tp, v, tb in self.exceptions:
            buffer.insert(
                buffer.get_end_iter(),
                f"Time since application startup: {datetime.timedelta(seconds=int(ts - START_TIME))}\n",
            )
            print_exception(tp, v, tb)

    def excepthook(
        self,
        exc_type: type[BaseException],
        exc_value: BaseException,
        exc_traceback: TracebackType | None,
    ):
        self.exceptions.append((time.time(), exc_type, exc_value, exc_traceback))
        log.exception(
            "Uncaught exception", exc_info=(exc_type, exc_value, exc_traceback)
        )
        self.update_text()
        if session := self.application.active_session:
            session.get_service("event_manager").handle(
                Notification(
                    gettext(
                        "An unexpected error occurred.\nPlease file an issue on GitHub and include information from Tools → Error Reports."
                    )
                )
            )



"""The definition for the profiles section of the toolbox."""

from gaphas.item import SE

from gaphor import UML
from gaphor.core import gettext
from gaphor.diagram.diagramtoolbox import ToolDef, ToolSection, new_item_factory
from gaphor.UML import diagramitems
from gaphor.UML.toolboxconfig import namespace_config


def metaclass_config(new_item):
    namespace_config(new_item)
    new_item.subject.name = new_item.diagram.gettext("Class")


profiles: ToolSection = ToolSection(
    gettext("Profiles"),
    (
        ToolDef(
            "toolbox-profile",
            gettext("Profile"),
            "gaphor-profile-symbolic",
            "r",
            new_item_factory(
                diagramitems.PackageItem,
                UML.Profile,
                config_func=namespace_config,
            ),
            handle_index=SE,
        ),
        ToolDef(
            "toolbox-metaclass",
            gettext("Metaclass"),
            "gaphor-metaclass-symbolic",
            "m",
            new_item_factory(
                diagramitems.ClassItem, UML.Class, config_func=metaclass_config
            ),
            handle_index=SE,
        ),
        ToolDef(
            "toolbox-stereotype",
            gettext("Stereotype"),
            "gaphor-stereotype-symbolic",
            "z",
            new_item_factory(
                diagramitems.ClassItem,
                UML.Stereotype,
                config_func=namespace_config,
            ),
            handle_index=SE,
        ),
        ToolDef(
            "toolbox-extension",
            gettext("Extension"),
            "gaphor-extension-symbolic",
            "<Shift>E",
            new_item_factory(diagramitems.ExtensionItem),
        ),
        ToolDef(
            "toolbox-import",
            gettext("Import"),
            "gaphor-import-symbolic",
            "<Shift>M",
            new_item_factory(diagramitems.PackageImportItem),
            handle_index=0,
        ),
    ),
)



import argparse
import logging
import os
import sys

from gaphor.application import distribution
from gaphor.entrypoint import initialize
from gaphor.plugins import default_plugin_path, enable_plugins

LOG_FORMAT = "%(name)s %(levelname)s %(message)s"


def main(argv=None) -> int:
    """Start Gaphor from the command line."""

    if argv is None:
        argv = sys.argv
    logging_config()

    with enable_plugins(default_plugin_path()):
        commands: dict[str, argparse.ArgumentParser] = initialize("gaphor.argparsers")

        args = parse_args(argv[1:], commands)

        return run_profiler(args) if args.profiler else args.command(args)  # type: ignore[no-any-return]


def run_profiler(args):
    import cProfile
    import pstats

    with cProfile.Profile() as profile:
        exit_code: int = profile.runcall(args.command, args)

    profile_stats = pstats.Stats(profile)
    profile_stats.strip_dirs().sort_stats("time").print_stats(50)
    return exit_code


def parse_args(args: list[str], commands: dict[str, argparse.ArgumentParser]):
    defaults = default_parser()
    parser = argparse.ArgumentParser(
        description="Gaphor is the simple modeling tool.",
        epilog="Thank you for using Gaphor <https://gaphor.org>.",
        parents=[version_parser()],
    )
    subparsers = parser.add_subparsers(
        title="subcommands (default: gui)",
        description=f"Get help for commands with {prog()} COMMAND --help.",
    )

    active_parser = parser
    for name, cmd_parser in commands.items():
        sp = subparsers.add_parser(
            name,
            description=cmd_parser.description,
            help=(cmd_parser.description or "").lower().rstrip("."),
            parents=[defaults, cmd_parser],
            add_help=False,
        )

        # Special case: fall back to gui subcommand if none is provided
        if name == "gui" and not (args and args[0] in commands):
            # Workaround: show toplevel help on the gui subcommand
            sp.format_help = parser.format_help  # type: ignore[method-assign]
            active_parser = sp

    return active_parser.parse_args(args)


def version_parser():
    parser = argparse.ArgumentParser(add_help=False)
    parser.add_argument(
        "-V", "--version", help="print version and exit", nargs=0, action=VersionAction
    )
    return parser


def default_parser():
    parser = argparse.ArgumentParser(add_help=False)

    loglevel = parser.add_mutually_exclusive_group()
    loglevel.add_argument(
        "-v",
        "--verbose",
        help="enable verbose logging",
        nargs=0,
        action=LogLevelAction,
        const=logging.DEBUG,
    )
    loglevel.add_argument(
        "-q",
        "--quiet",
        help="only show warning and error logging",
        nargs=0,
        action=LogLevelAction,
        const=logging.WARNING,
    )
    parser.add_argument(
        "--profiler", help="run in profiler (cProfile)", action="store_true"
    )
    return parser


def gui_parser():
    def run(args) -> int:
        # Only now import the UI module
        import gaphor.ui

        # Recreate a command line for our GTK gui
        run_argv = [sys.argv[0]]
        if args.self_test:
            run_argv += ["--self-test"]
        if args.gapplication_service:
            run_argv += ["--gapplication-service"]
        run_argv.extend(args.model)

        return gaphor.ui.run(run_argv)

    parser = argparse.ArgumentParser(
        description="Launch the GUI.", parents=[version_parser()]
    )

    group = parser.add_argument_group("options (no command provided)")
    group.add_argument(
        "--self-test", help="run self test and exit", action="store_true"
    )
    group.add_argument("--gapplication-service", action="store_true")
    group.add_argument("model", nargs="*", help="model file(s) to load")
    parser.set_defaults(command=run)
    return parser


def exec_parser():
    def execute_script(args) -> int:
        import runpy

        script = args.script
        runpy.run_path(script, run_name="__main__")
        return 0

    parser = argparse.ArgumentParser(
        description="Execute a python script from within Gaphor."
    )
    parser.add_argument("script", help="execute a script file and exit")
    parser.set_defaults(command=execute_script)
    return parser


def logging_config(level=logging.INFO):
    if level <= logging.DEBUG:
        logging.basicConfig(level=logging.INFO, format=LOG_FORMAT, force=True)
        logging.getLogger("gaphor").setLevel(logging.DEBUG)
    elif level >= logging.WARNING:
        logging.basicConfig(level=logging.WARNING, format=LOG_FORMAT, force=True)
    else:
        logging.basicConfig(level=logging.INFO, format=LOG_FORMAT, force=True)


class VersionAction(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        print(f"Gaphor {distribution().version}")
        parser.exit()


class LogLevelAction(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        logging_config(self.const)


def prog():
    return os.path.basename(sys.argv[0])



from __future__ import annotations

import sys
from typing import NamedTuple

from gi.repository import Gio, GLib, Gtk

from gaphor.abc import ActionProvider
from gaphor.action import action


def apply_application_actions(component_registry, gtk_app):
    scope = "app"
    for _name, provider in component_registry.all(ActionProvider):
        for attrname, act in iter_actions(provider, scope):
            a = create_gio_action(act, provider, attrname)
            gtk_app.add_action(a)
            if act.shortcut:
                gtk_app.set_accels_for_action(
                    f"{scope}.{act.name}",
                    [_platform_specific(s) for s in act.shortcuts],
                )
    return gtk_app


class ActionGroup(NamedTuple):
    actions: Gio.SimpleActionGroup
    shortcuts: Gio.ListModel


def window_action_group(component_registry) -> ActionGroup:
    action_group = Gio.SimpleActionGroup.new()
    store = Gio.ListStore.new(Gtk.Shortcut)
    scope = "win"
    for _name, provider in component_registry.all(ActionProvider):
        for attrname, act in iter_actions(provider, scope):
            a = create_gio_action(act, provider, attrname)
            action_group.add_action(a)
            for shortcut in act.shortcuts:
                store.append(
                    _new_shortcut(_platform_specific(shortcut), act.detailed_name)
                )
    return ActionGroup(actions=action_group, shortcuts=store)


def create_action_group(provider, scope) -> ActionGroup:
    action_group = Gio.SimpleActionGroup.new()
    store = Gio.ListStore.new(Gtk.Shortcut)
    for attrname, act in iter_actions(provider, scope):
        a = create_gio_action(act, provider, attrname)
        action_group.add_action(a)
        for shortcut in act.shortcuts:
            store.append(_new_shortcut(_platform_specific(shortcut), act.detailed_name))
    return ActionGroup(actions=action_group, shortcuts=store)


def _platform_specific(shortcut):
    return shortcut.replace(
        "<Primary>", "<Meta>" if sys.platform == "darwin" else "<Control>"
    )


def _new_shortcut(shortcut, detailed_name):
    return Gtk.Shortcut.new(
        trigger=Gtk.ShortcutTrigger.parse_string(shortcut),
        action=Gtk.NamedAction.new(detailed_name),
    )


def create_gio_action(act, provider, attrname):
    if act.state is not None:
        state = act.state(provider) if callable(act.state) else act.state
        a = Gio.SimpleAction.new_stateful(
            act.name, as_variant_type(act.arg_type), to_variant(state)
        )
        a.connect("change-state", _action_activate, provider, attrname)
    else:
        a = Gio.SimpleAction.new(act.name, as_variant_type(act.arg_type))
        a.connect("activate", _action_activate, provider, attrname)
    return a


def iter_actions(provider, scope):
    provider_class = type(provider)
    for attrname in dir(provider_class):
        method = getattr(provider_class, attrname)
        act: action | None = getattr(method, "__action__", None)
        if act and act.scope == scope:
            yield (attrname, act)


_GVARIANT_TYPE_STR = GLib.VariantType.new("s")
_GVARIANT_TYPE_INT = GLib.VariantType.new("i")
_GVARIANT_TYPE_BOOL = GLib.VariantType.new("b")


def as_variant_type(t):
    if t is None:
        return None
    elif t is str:
        return _GVARIANT_TYPE_STR
    elif t is int:
        return _GVARIANT_TYPE_INT
    elif t is bool:
        return _GVARIANT_TYPE_BOOL
    else:
        raise ValueError(f"No GVariantType declared for Python type {t}")


def to_variant(v):
    if v is None:
        return None
    elif isinstance(v, str):
        return GLib.Variant.new_string(v)
    elif isinstance(v, bool):
        return GLib.Variant.new_boolean(v)
    elif isinstance(v, int):
        return GLib.Variant.new_int32(v)
    else:
        raise ValueError(
            f"No GVariant mapping declared for Python value {v} of type {type(v)}"
        )


def from_variant(v):
    if v is None:
        return None
    elif _GVARIANT_TYPE_STR.equal(v.get_type()):
        return v.get_string()
    elif _GVARIANT_TYPE_INT.equal(v.get_type()):
        return v.get_int32()
    elif _GVARIANT_TYPE_BOOL.equal(v.get_type()):
        return v.get_boolean()
    else:
        raise ValueError(f"No Python mapping declared for GVariant value {v}")


def _action_activate(action, param, obj, name):
    method = getattr(obj, name)
    if param is not None:
        method(from_variant(param))
        if action.get_state_type():
            action.set_state(param)
    else:
        method()



# ruff: noqa: F401

from gaphor.UML.actions import (
    actionseditors,
    actionspropertypages,
    activitymove,
    activitypropertypage,
    partitionpage,
)
from gaphor.UML.classes import (
    associationpropertypages,
    classeseditors,
    classespropertypages,
    dependencypropertypages,
    enumerationpropertypages,
)
from gaphor.UML.interactions import interactionspropertypages
from gaphor.UML.profiles import metaclasspropertypage, stereotypepropertypages
from gaphor.UML.states import dropzone, propertypages



"""Trivial drawing aids (box, line, ellipse)."""

from gaphor.diagram.presentation import ElementPresentation, LinePresentation
from gaphor.diagram.shapes import Box as BoxShape
from gaphor.diagram.shapes import draw_border, draw_ellipse


class Line(LinePresentation):
    def __init__(self, diagram, id=None):
        super().__init__(diagram=diagram, id=id)
        self._handles[0].connectable = False
        self._handles[-1].connectable = False


class Box(ElementPresentation):
    def __init__(self, diagram, id=None):
        super().__init__(diagram=diagram, id=id)
        self.shape = BoxShape(
            draw=draw_border,
        )


class Ellipse(ElementPresentation):
    def __init__(self, diagram, id=None):
        super().__init__(diagram=diagram, id=id)
        self.shape = BoxShape(
            draw=draw_ellipse,
        )
